#ifndef _HARK__BEAMFORMING__MATRIX_H__INCLUDED_
#define _HARK__BEAMFORMING__MATRIX_H__INCLUDED_ 1

#include <complex>
#include <vector>

namespace BF
{
    typedef std::complex<double> Complex;

    class Matrix
    {
    public:
	Matrix();
	Matrix(int rows, int cols);
	Matrix(const Matrix &src);
	~Matrix();

	Matrix &operator = (const Matrix &src);

    public:
	int getRowCount() const;
	int getColumnCount() const;

	void setElementAt(int row, int col, double rvalue, double ivalue = 0);
	void setElementAt(int row, int col, const Complex &value);
	const Complex &getElementAt(int row, int col) const;

	const Matrix getColumn(int col) const;
	void setColumn(int col, const Matrix &value);

	const Matrix getRow(int row) const;
	void setRow(int row, const Matrix &value);

	const Matrix getSubmatrix(int row, int col, int rows, int cols) const;
	void setSubmatrix(int row, int col, const Matrix &m);

	Complex scalar() const; /* 行列が 1x1 の時、スカラー値に変換して返す */

	const Matrix  pseudoInverse() const;
	const Matrix  Inverse() const;

	/* 特異値分解を行う。
	 *   [U, S, V] = svd(this) として、U, S, V をそれぞれ引数 u, s, v に収めて返す。
	 *
	 * 特異値分解の定義から、
	 *   u.getRowCount() == u.getColumnCount() == this->getRowCount()
	 *   s.getRowCount() == this->getRowCount(), s.getColumnCount() == this->getColumnCount()
	 *   v.getRowCount() == v.getColumnCount() == this->getColumnCount()
	 * である。
	 */
	void singularValueDecomposition(Matrix &u, Matrix &s, Matrix &v) const;

	/* this はエルミート行列であると仮定し、固有値・固有ベクトルを求める。
	 * this がエルミート行列でない場合、得られる値は正しくない。
	 */
	void eigenValues(Matrix &eigenValues, Matrix &eigenVectors) const;

	/* this はエルミート行列、mat は正定値エルミート行列であると仮
	 * 定し、一般固有値・固有ベクトルを求める。
	 *
	 * this がエルミート行列でない場合、あるいは mat が正定値エルミー
	 * ト行列でない場合、得られる値は正しくない。
	 */
	void generalizedEigenValues(const Matrix &mat, Matrix &eigenValues, Matrix &eigenVectors) const;

	/* this, mat はエルミート行列であると仮
	 * 定し、一般化特異値・特異ベクトルを求める。
	 */
        void generalizedSingularValues(const Matrix &K, Matrix &u, Matrix &s, Matrix &v) const;

	Complex norm() const;
	Complex frobeniusNorm() const;

	const Matrix  absoluteValues() const;   /* 各要素を絶対値化した実行列を返す */
	const Matrix  diagonalElements() const; /* 対角成分のみを抽出したベクトルを返す */
	const Matrix  diagonalMatrix() const;   /* ベクトルから対角行列を作成する */
	Complex       getMaximumValue() const;
	int           getMaximumValueIndexInRow(int row) const;
	int           getMaximumValueIndexInColumn(int col) const;
	Complex       getElementSum() const;

	Matrix       &transpose(bool conj = true);
	const Matrix  transposed(bool conj = true) const;

	Matrix       &conjugate();
	const Matrix  conjugated() const;

	Matrix       &multiply(const Matrix &rhs);
	const Matrix  multiplied(const Matrix &rhs) const;
	
	Matrix       &add(const Matrix &rhs);
	const Matrix  added(const Matrix &rhs) const;
	
	Matrix       &subtract(const Matrix &rhs);
	const Matrix  subtracted(const Matrix &rhs) const;

	const Matrix  elementwiseMultiplied(const Matrix &rhs) const;
	const Matrix  elementwiseFMultiplied(const Matrix &rhs) const;
	const Matrix  elementwiseDivided(const Matrix &rhs) const;
	const Matrix  elementwiseNorm() const;
	const Matrix  elementwiseAbs() const;
	const Matrix  elementwiseSqrt() const;

	static const Matrix I(int size);
	static const Matrix O(int rows, int cols);
	
	void swap(Matrix &src);

	void dump(FILE *stream) const;
	void dump(FILE *stream, const char *format, ...) const;

    private:
	int rows, cols;
	Complex *values;

    };

    const Matrix operator + (const Matrix  &lhs, const Matrix  &rhs);
    const Matrix operator - (const Matrix  &lhs, const Matrix  &rhs);
    const Matrix operator * (const Matrix  &lhs, const Matrix  &rhs);
    const Matrix operator * (const Complex &lhs, const Matrix  &rhs);
    const Matrix operator * (const Matrix  &lhs, const Complex &rhs);

    class Location
    {
    public:
	Location(double _x, double _y, double _z);
    
    public:
	double x, y, z;

    };

    class MicLocation : public Location
    {
    public:
	MicLocation(double _x, double _y, double _z);

    };

    class SourceLocation : public Location
    {
    public:
	SourceLocation(double _x, double _y, double _z, int sourceID, bool isSignal);

    public:
	int  sourceID;
	bool isSignal;
    };

    class TransformFunctionDatabase
    {
    public:
	std::vector<SourceLocation> sourceLocations;
	std::vector<MicLocation>    micLocations;
	std::vector<Matrix>         transformFunctions; /* size_is(numFreq), elemnent_is(Matrix(numMic, numSrc)) */

    public:
	int findNearestSourceIndex(const SourceLocation &sourceLocation) const;
	int findNearestSourceIndex(const MicLocation    &micLocation) const;

    };
}

#include <algorithm>

namespace std
{
    template<> void swap<BF::Matrix>(BF::Matrix &a, BF::Matrix &b);
}

#endif
